import zipfile
import io
import os
from typing import Optional, Dict, Any
from .base import BaseScanner, ScanResult, IssueSeverity
from .pickle_scanner import PickleScanner

class PyTorchZipScanner(BaseScanner):
    """Scanner for PyTorch Zip-based model files (.pt, .pth)"""
    name = "pytorch_zip"
    description = "Scans PyTorch model files for suspicious code in embedded pickles"
    supported_extensions = [".pt", ".pth"]
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        super().__init__(config)
        # Initialize a pickle scanner for embedded pickles
        self.pickle_scanner = PickleScanner(config)
        
    @classmethod
    def can_handle(cls, path: str) -> bool:
        """Check if this scanner can handle the given path"""
        if not os.path.isfile(path):
            return False
            
        # Check file extension
        ext = os.path.splitext(path)[1].lower()
        if ext not in cls.supported_extensions:
            return False
            
        # Verify it's a zip file
        try:
            with zipfile.ZipFile(path, "r") as _:
                pass
            return True
        except zipfile.BadZipFile:
            return False
        except Exception:
            return False
    
    def scan(self, path: str) -> ScanResult:
        """Scan a PyTorch model file for suspicious code"""
        # Check if path is valid
        path_check_result = self._check_path(path)
        if path_check_result:
            return path_check_result
            
        result = self._create_result()
        file_size = self.get_file_size(path)
        result.metadata["file_size"] = file_size
        
        try:
            # Store the file path for use in issue locations
            self.current_file_path = path
            
            with zipfile.ZipFile(path, "r") as z:
                pickle_files = [name for name in z.namelist() if name.endswith(".pkl")]
                result.metadata["pickle_files"] = pickle_files
                
                # Track number of bytes scanned
                bytes_scanned = 0
                
                # Scan each pickle file
                for name in pickle_files:
                    data = z.read(name)
                    bytes_scanned += len(data)
                    
                    file_like = io.BytesIO(data)
                    # Use the pickle scanner directly
                    sub_result = self.pickle_scanner._scan_pickle_bytes(file_like, len(data))
                    
                    # Include the pickle filename in each issue
                    for issue in sub_result.issues:
                        if issue.details:
                            issue.details["pickle_filename"] = name
                        else:
                            issue.details = {"pickle_filename": name}
                        
                        # Update location to include the main file path
                        if not issue.location:
                            issue.location = f"{path}:{name}"
                        elif "pos" in issue.location:
                            # If it's a position from the pickle scanner, prepend the file path
                            issue.location = f"{path}:{name} {issue.location}"
                    
                    # Merge results
                    result.merge(sub_result)
                
                # Check for other suspicious files
                for name in z.namelist():
                    # Check for Python code files
                    if name.endswith(".py"):
                        result.add_issue(
                            f"Python code file found in PyTorch model: {name}",
                            severity=IssueSeverity.WARNING,
                            location=f"{path}:{name}",
                            details={"file": name}
                        )
                    # Check for shell scripts or other executable files
                    elif name.endswith((".sh", ".bash", ".cmd", ".exe")):
                        result.add_issue(
                            f"Executable file found in PyTorch model: {name}",
                            severity=IssueSeverity.ERROR,
                            location=f"{path}:{name}",
                            details={"file": name}
                        )
                
                result.bytes_scanned = bytes_scanned
                
        except zipfile.BadZipFile:
            result.add_issue(
                f"Not a valid zip file: {path}",
                severity=IssueSeverity.ERROR,
                location=path,
                details={"path": path}
            )
            result.finish(success=False)
            return result
        except Exception as e:
            result.add_issue(
                f"Error scanning PyTorch zip file: {str(e)}",
                severity=IssueSeverity.ERROR,
                location=path,
                details={"exception": str(e), "exception_type": type(e).__name__}
            )
            result.finish(success=False)
            return result
            
        result.finish(success=True)
        return result
